from time import sleep
from functools import partial
from posted.base import MsgBrokerBase, NoMsg

EXISTING_CHANNEL = "existing-channel"

MESSAGES = [
    "some string",
    123,
    3.14,
    {"a": 1, "b": 2, "c": 3},
    None,
]
CHANNEL_NAMES = [
    EXISTING_CHANNEL,
    # 'new-channel',
]


def gen_test_mk_msg_broker_args():
    for msg in MESSAGES:
        for queue in CHANNEL_NAMES:
            yield (msg, queue)


def base_test_on_demand_consumption(broker: MsgBrokerBase, message, channel):
    assert broker.read(channel) is NoMsg  # queue is empty
    broker.write(channel, message)
    assert broker.read(channel) == message
    assert broker.read(channel) is NoMsg  # queue is empty again


def base_test_reactive_consumption(broker: MsgBrokerBase, message, channel):
    CONTENT = "content"
    sub_msg = {}

    wait = partial(sleep, 0.2)  # sleep for 200ms

    def callback(msg):
        sub_msg[CONTENT] = msg

    # subscribe, write, and check
    broker.subscribe(channel, callback)
    wait()
    broker.write(channel, message)
    wait()
    assert sub_msg.get(CONTENT) == message

    # write, subscribe, and check
    broker.write(channel, message)
    wait()
    broker.subscribe(channel, callback)
    wait()
    assert sub_msg.get(CONTENT) == message

    # unsubscribe, write, and check
    sub_msg = {}
    broker.unsubscribe(channel)
    wait()
    broker.write(channel, message)
    wait()
    assert CONTENT not in sub_msg



import subprocess
from time import sleep


def run_cmd(cmd: str, run_in_background: bool = False):
    cmd_args = cmd.split(" ")

    if run_in_background:
        process = subprocess.Popen(cmd_args, stdout=subprocess.DEVNULL)
        return process

    result = subprocess.run(cmd_args, stdout=subprocess.PIPE)
    return result.stdout.decode() if result.stdout else None


def service_is_running(service_container_name: str):
    output = run_cmd(f"docker ps -q -f name={service_container_name}")
    return bool(output)


def start_service(
    service_container_name: str, docker_compose_filepath: str, *, wait: float = 2.0
):
    """
    Start a docker service container.

    :param service_container_name: The name of the service container.
    :param docker_compose_filepath: The path to the docker-compose file.
    :param wait: The time to wait after starting the container, in seconds.
    """
    print("\n")
    print(f"========================== Running container ==========================")
    run_cmd(
        f"docker-compose -f {docker_compose_filepath} -p {service_container_name} up -d"
    )
    sleep(wait)
    print(f"========================== Container running ==========================")


def stop_service(service_container_name: str, docker_compose_filepath: str):
    """
    Stop a docker service container.

    :param service_container_name: The name of the service container.
    :param docker_compose_filepath: The path to the docker-compose file.
    """
    print("\n")
    print(f"========================== Stopping container ==========================")
    run_cmd(
        f"docker-compose -f {docker_compose_filepath} -p {service_container_name} stop"
    )
    print(f"========================== Container stopped ===========================")



"""
Base module for the `posted` package.

This module contains the base class for message brokers, 'MsgBrokerBase', which defines
the interface for message brokers.
"""

from abc import ABC
from functools import partial
import json
from typing import Any, Callable, Mapping
from i2.util import mk_sentinel
import concurrent.futures


_mk_sentinel = partial(
    mk_sentinel, boolean_value=False, repr_=lambda x: x.__name__, module=__name__
)
NoMsg = _mk_sentinel("NoMsg")


def _dflt_encoder(v: Any) -> bytes:
    return json.dumps(v).encode("utf-8")


def _dflt_decoder(v: bytes) -> Any:
    return json.loads(v.decode("utf-8"))


class MsgBrokerBase(ABC):
    """
    Base class for message brokers. Subclasses must implement the 'write', 'read',
    'subscribe', and 'unsubscribe' methods.
    """

    # _encoder: Callable[[Any], bytes]
    # _decoder: Callable[[bytes], Any]
    _config: Mapping[str, Any]

    def __init__(
        self,
        *,
        encoder: Callable[[Any], bytes] = None,
        decoder: Callable[[bytes], Any] = None,
        **kwargs
    ):
        """
        Initialize the message broker.

        :param encoder: The encoder to use for encoding messages. By default, the
            `json` module is used to dump messages to JSON.
        :param decoder: The decoder to use for decoding messages. By default, the
            `json` module is used to load messages from JSON.
        :param kwargs: Additional configuration parameters, specific to the target
            infrastructure.
        """
        self._encoder = encoder or _dflt_encoder
        self._decoder = decoder or _dflt_decoder
        self._config = kwargs

        self._subscriptions = {}
        self._executor = concurrent.futures.ThreadPoolExecutor()

    def write(self, channel: str, message: Any):
        """
        Write a message to the channel.

        :param channel: The channel to write to.
        :param message: The message to write.
        """
        raise NotImplementedError("The 'write' method must be implemented in subclass.")

    def read(self, channel: str):
        """
        Read a message from the channel.

        :param channel: The channel to read from.
        :return: The message read from the channel.
        """
        raise NotImplementedError("The 'read' method must be implemented in subclass.")

    def subscribe(self, channel: str, callback: Callable[[Any], None]):
        """
        Subscribe to a channel.

        :param channel: The channel to subscribe to.
        :param callback: The callback to call when a message is received on the channel.
        """
        raise NotImplementedError(
            "The 'subscribe' method must be implemented in subclass."
        )

    def unsubscribe(self, channel: str):
        """
        Unsubscribe from a channel.

        :param channel: The channel to unsubscribe from.
        """
        raise NotImplementedError(
            "The 'unsubscribe' method must be implemented in subclass."
        )
